#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <qdos.h>
#include <sys/stat.h>
#include "tar.h"

char _prog_name[] = "tar";
char _copyright[] = __DATE__;
char _version[] = "0.05";
char *  _endmsg = NULL;

long _stack = 8*1024;

struct WINDOWDEF _condetails =  { 208, 1, 0, 7, 512, 256, 0, 0};

void (*_consetup)() = consetup_title;

int custom_expand (char * param, char ***argvptr, int *argcptr);

int (*_cmdwildcard)() = custom_expand;

/* custom cmdexpand: also expands directory names */

#define FILEBUF_INIT    1024        /* Initial allocation size for buffer */
#define FILEBUF_INCR    1024        /* Increment size for buffer */

int custom_expand (char * param, char ***argvptr, int *argcptr)
{
    int     count,sl;
    size_t  bufsize;
    char    *filenamebuf;
    char    *ptr,*safeptr;

    /*
     *  Check to see if we should do wild card expansion.
     *  We only perform wildcard expansion if the parameter
     *  was not a string and if it contains one of the
     *  wild card characters.
     *
     *  We also do not expand any option that starts with '-'
     *  as we then assume that it is a unix stylew option.
     */
    if ((*param == '-') ||  (strpbrk(param,"*?") == NULL) ) {
        return 0;
    }

    if ((filenamebuf = malloc (bufsize = FILEBUF_INIT)) == NULL) {
        return -1;
    }
TRYAGAIN:
    count = getfnl(param, filenamebuf, bufsize, QDR_ALL);
    if (count == -1  && errno == ENOMEM) {
        /*
         *  We have overflowed the buffer, so we try
         *  to get a bigger buffer and try again.
         */
        bufsize += FILEBUF_INCR;
        if ((filenamebuf = realloc (filenamebuf, bufsize)) == NULL) {
            return -1;
        } else {
            goto TRYAGAIN;
        }
    }
    /*
     *  If no files were found, then return unexpanded.
     */
    if (count == 0) {
        free (filenamebuf);
        return 0;
    }
    /*
     *  Files were found, so add these to the list instead
     *  of the original parameter typed by the user.
     */
    for ( ptr=filenamebuf ; count > 0 ; count -- ) {
        *argvptr = (char **)realloc (*argvptr, (size_t)(((*argcptr) + 2) * sizeof (char *)));
        safeptr=(char *)malloc((size_t)(sl=strlen(ptr)+1));
        if (safeptr == NULL || *argvptr == NULL) {
            return -1;
        }
        (void) memcpy(safeptr,ptr,(size_t)sl);
        (*argvptr)[*argcptr] = safeptr;
        *argcptr += 1;
        ptr += sl;
    }
    free (filenamebuf);
    return *argcptr;
}

extern char odd[];

static jobid_t chowner(chanid_t chan)
{
    extern char *_sys_var;
    char *scht;
    long *cdb;
    long jid;
    
    scht = *((char **)(_sys_var + 0x78));
    cdb = *(long **)((long *)scht  + (chan & 0xffff));
    jid = *(cdb + 2);
    return jid;
}

extern short qlwait;
extern short qlflag;

void qexit (void)
{
    jobid_t me,you;
    
    if(*odd)
    {
        chdir(odd);
    }
    
    me = getpid();
    you = chowner(getchid(0));
    
    if(me == you)
    {
        if(isatty(0) && isatty(2) && qlwait)
        {
            char c = 0;
            fputs("Press a key to exit", stderr);
            if((io_fbyte(getchid(0), qlwait, &c) == 0) && c == 27)
            {
                io_fbyte(getchid(0), -1, &c);
            }
        }
    }
}
typedef struct __dlp__ 
{
    struct __dlp__ *next;
    char *dln;
    size_t dll;
} DLP;

char *cvtUNIX2SMS(char *name)
{
    char *p;
    for(p = name; *p; p++)
    {
        if(*p == '/' || (*p == '.' && (qlflag & 4))) *p = '_';
    }
    return name;
}

void SetHeader(char *p, union record *h)
{
    int fd;

    strcpy(h->header.magic_cookie, "SMSQDOSTar");    
    fd = open(p, O_RDONLY); /* use c i/o to use def dir */
    if (fd >  0)
    {
        struct qdirect s;
        fs_headr(getchid(fd), -1, (char *)&s, (int) sizeof(s));
        close(fd);
        h->header.d_length = s.d_length;
        h->header.d_access = s.d_access;
        h->header.d_type = s.d_type;
        h->header.d_datalen = s.d_datalen;
        h->header.d_reserved = s.d_reserved;
    }
}        

void SetSMSStuff(int fd, union record *h)
{
    if((strcmp(h->header.magic_cookie, "SMSQDOSTar") == 0) && h->header.d_type)
    {
        int ch = getchid(fd);
        fs_heads(ch, (timeout_t)-1, &h->header.d_length, 14);
    }
}
static DLP *AddDent(char *s, DLP *dlp, int n)
{
    DLP *d;
    d = malloc(sizeof(DLP));
    d->dln = strdup(s);
    d->dll = strlen(s);
    d->next = dlp;
    dlp = d;
    if(n)
    {
        s[n] = '/';
    }
    return dlp;
}

char *cvtSMS2UNIX(char *name, char *newname)
{
    char wd[36];
    char *p;
    char *s;
    int n;
    static DLP *dlp = NULL;
    
    getcwd(wd, sizeof(wd));
    if(strstr(name, wd))
    {
        s = name + strlen(wd);
    }
    else
    {
        s = name;
    }
    
    n = strlen(s);

    if(n == 0)
    {
        strcpy(newname, "./");
    }
    else
    {
        s = strcpy(newname, s);
        if(s[n-1] == '_')
        {
            dlp = AddDent(s, dlp, n-1);
        }
        else
        {
            char *q;
            int j;

            /* First check for hidden dirs */

            if(dlp == NULL)
            {
                char *r;
                int sts;
                struct stat st;
                
                for(r = s; *r; r++)
                {
                    if(*r == '_')
                    {
                        *r = 0;
                        sts = stat(s, &st);
                        *r = '_';
                        
                        if(sts ==0 && (st.st_mode & S_IFDIR))
                        {
                            char cc = *(r+1);
                            *(r+1) = 0;
                            dlp = AddDent(s, dlp, 0);
                            *(r+1) = cc;
                        }
                    }
                }
            }
            
            while(dlp)
            {
                if(strncmp(s, dlp->dln, dlp->dll) == 0) /* matches this */
                {
                    for(j = 0; j < dlp->dll; j++)
                    {
                        if(s[j] == '_') s[j] = '/';
                    }
                    break;
                }
                else
                {
                    DLP *z;
                    {
                        z = dlp;
                        dlp = dlp->next;
                        free(z->dln);
                        free(z);
                    }
                }
            }
            if((q = strrchr(s+j, '_')))
            {
                *q = '.';
            }
        }
    }
    return newname;
}
